"""
Run via python -m unittest in parent folder

"""


from unittest import TestCase

from caselessmapping import CaselessMapping, CaselessMutableMapping


class TestCaselessMapping(TestCase):

    def test_null_init(self):
        self.assertRaises(TypeError, CaselessMapping)

    def test_int_keys(self):
        data = {1: 10, 2: 20, 3:30}
        cd = CaselessMapping(data)
        self.assertEqual(cd, data)

    def test_str_keys(self):
        data = {'a': 'A', 'b': 'B', 'c': 'C'}
        cd = CaselessMapping(data)
        self.assertEqual(cd, data)

    def test_matching_case_retreive(self):
        data = {'a': 'A', 'B': 'B', 'c': 'C'}
        cd = CaselessMapping(data)
        self.assertEqual(cd['a'], data['a'])
        self.assertEqual(cd['B'], data['B'])

    def test_caseless_retreive(self):
        data = {'a': 'A', 'B': 'B', 'c': 'C'}
        cd = CaselessMapping(data)
        self.assertEqual(cd['A'], data['a'])
        self.assertEqual(cd['b'], data['B'])

    def test_len(self):
        data = {'a': 'A', 'B': 'B', 'c': 'C'}
        cd = CaselessMapping(data)
        self.assertEqual(len(cd), 3)

    def test_iter(self):
        data = {'a': 'A', 'B': 'B', 'c': 'C'}
        cd = CaselessMapping(data)
        orig_tups = list(data.items())
        test_tups = list(cd.items())
        self.assertListEqual(orig_tups, test_tups)

    def test_raw_name(self):
        data = {'a': 'A', 'B': 'B', 'c': 'C'}
        cd = CaselessMapping(data)
        self.assertEqual(cd.get_raw_key_name('b'), 'B')


class TestCaselessMutableMapping(TestCase):

    def test_null_init(self):
        cd = CaselessMutableMapping()
        self.assertEqual(cd, {})

    def test_int_keys(self):
        data = {1: 10, 2: 20, 3:30}
        cd = CaselessMutableMapping(data)
        self.assertEqual(cd, data)

    def test_str_keys(self):
        data = {'a': 'A', 'b': 'B', 'c': 'C'}
        cd = CaselessMutableMapping(data)
        self.assertEqual(cd, data)

    def test_matching_case_retreive(self):
        data = {'a': 'A', 'B': 'B', 'c': 'C'}
        cd = CaselessMutableMapping(data)
        self.assertEqual(cd['a'], data['a'])
        self.assertEqual(cd['B'], data['B'])

    def test_caseless_retreive(self):
        data = {'a': 'A', 'B': 'B', 'c': 'C'}
        cd = CaselessMutableMapping(data)
        self.assertEqual(cd['A'], data['a'])
        self.assertEqual(cd['b'], data['B'])

    def test_len(self):
        data = {'a': 'A', 'B': 'B', 'c': 'C'}
        cd = CaselessMutableMapping(data)
        self.assertEqual(len(cd), 3)

    def test_iter(self):
        data = {'a': 'A', 'B': 'B', 'c': 'C'}
        cd = CaselessMutableMapping(data)
        orig_tups = list(data.items())
        test_tups = list(cd.items())
        self.assertListEqual(orig_tups, test_tups)

    def test_raw_name(self):
        data = {'a': 'A', 'B': 'B', 'c': 'C'}
        cd = CaselessMutableMapping(data)
        self.assertEqual(cd.get_raw_key_name('b'), 'B')

    def test_add_item_str_key(self):
       cd = CaselessMutableMapping()
       cd['x'] = 'X'
       self.assertEqual(cd, {'x': 'X'})

    def test_add_item_int_key(self):
       cd = CaselessMutableMapping()
       cd[1] = 'X'
       self.assertEqual(cd, {1: 'X'})

    def test_del_item_str_key(self):
        data = {'x': 'X', 'y': 'Y'}
        cd = CaselessMutableMapping(data)
        del cd['x']
        self.assertEqual(cd, {'y': 'Y'})

    def test_del_item_int_key(self):
        data = {1: 'X', 2: 'Y'}
        cd = CaselessMutableMapping(data)
        del cd[1]
        self.assertEqual(cd, {2: 'Y'})




from collections.abc import Mapping, MutableMapping


class CaselessMapping(Mapping):

    def __init__(self, init):
        init_dict = dict(init)
        self._data = {self.norm_key(key): (key, value) for key, value in init_dict.items()}

    @staticmethod
    def norm_key(key):
        if isinstance(key, str):
            return key.lower()
        else:
            return key

    def __getitem__(self, key):
        return self._data[self.norm_key(key)][1]

    def __len__(self):
        return len(self._data)

    def __iter__(self):
        for key in self._data.keys():
            yield self._data[key][0]

    def get_raw_key_name(self, key):
        return self._data[self.norm_key(key)][0]


class CaselessMutableMapping(MutableMapping, CaselessMapping):

    def __init__(self, init=None):
        super().__init__(init if init is not None else {})

    def __setitem__(self, key, value):
        self._data[self.norm_key(key)] = (key, value)

    def __delitem__(self, key):
        del self._data[self.norm_key(key)]





from collections.abc import MutableMapping


class EmptyDict(MutableMapping):

    def __init__(self, *args, **kwargs):
        super().__init__()

    def __getitem__(self, key, *args, **kwargs):
        raise KeyError(key)

    def __len__(self):
        return 0

    def __iter__(self):
        return iter([])

    def __setitem__(self, *args, **kwargs):
        pass

    def __delitem__(self, key, *args, **kwargs):
        raise KeyError(key)





class CastBase:
    """
    Identical to regular dict/list except on instantiation it performs a
    one-off recursive recasting of all contained items that have a type listed
    in cast_map to the mapped target type.

    Also automatically recasts any list or dict instances to CastList and
    CastDict respectively.

    Useful if you have stubborn objects that won't pickle or export to json for
    example.

    cast_map is a dict of conversion types and their recast targets, e.g.:

        {int: str, float: Decimal}

    Conversion types can be a tuple of types:

        {(int, float): str}

    """
    def __init__(self, *args, cast_map, **kwargs):
        self.cast_map = cast_map
        super().__init__(*args, **kwargs)
        self._recast()

    def _recast(self):
        raise NotImplemented()

    def _recast_item(self, item):
        # Try to recast it first, then recurse
        for src, dst in self.cast_map.items():
            if isinstance(item, src):
                item = dst(item)
        if isinstance(item, list):
            return CastList(item, cast_map=self.cast_map)
        elif isinstance(item, dict):
            return CastDict(item, cast_map=self.cast_map)
        else:
            return item


class CastDictMixin(CastBase):
    def _recast(self):
        for k, v in self.items():
            self[k] = self._recast_item(v)


class CastListMixin(CastBase):
    def _recast(self):
        for i, item in enumerate(self):
            self[i] = self._recast_item(item)


class CastDict(CastDictMixin, dict):
    pass


class CastList(CastListMixin, list):
    pass





"""
Run via python -m unittest in parent folder

"""


import sys
import unittest

from casttypes import *


class TestCastDict(unittest.TestCase):

    def test_init1(self):
        d = {1: 10, 2:20}
        cd = CastDict(d, cast_map={})
        self.assertDictEqual(cd, d)

    def test_init2(self):
        d = {1: 10, 2:20}
        cd = CastDict([(1, 10), (2, 20)], cast_map={})
        self.assertDictEqual(cd, d)

    def test_init3(self):
        d = {(1, 10): 100, (2, 20): 200}
        cd = CastDict(d, cast_map={})
        self.assertDictEqual(cd, d)

    def test_init4(self):
        d = {1: [1, 2, 3], 2: [2, 3, 4]}
        cd = CastDict(d, cast_map={})
        self.assertDictEqual(cd, d)

    def test_init5(self):
        d = {1: {10: ['a', 'b'], 20: ['A', 'B']}, '2': (5, 6)}
        cd = CastDict(d, cast_map={})
        self.assertDictEqual(cd, d)

    def test_cast1(self):
        d = {1: {10: 100}}
        cd = CastDict(d, cast_map={dict: dict})
        self.assertDictEqual(cd, d)

    def test_cast2(self):
        d = {1: {10: 100}}
        cd = CastDict(d, cast_map={list: list})
        self.assertDictEqual(cd, d)

    def test_cast3(self):
        d = {1: {10: 100}}
        cd = CastDict(d, cast_map={dict: list})
        self.assertDictEqual(cd, {1: [10]})

    def test_cast4(self):
        d = {1: [(1, 10), (2, 20)]}
        cd = CastDict(d, cast_map={list: dict})
        self.assertDictEqual(cd, {1: {1: 10, 2: 20}})

    def test_cast5(self):
        d = {'1': '10', '2': '20'}
        cd = CastDict(d, cast_map={str: int})
        self.assertDictEqual(cd, {'1': 10, '2': 20})

    def test_cast6(self):
        d = {1: (2, 3, 4)}
        cd = CastDict(d, cast_map={list: str})
        self.assertDictEqual(cd, d)

    def test_cast7(self):
        d = {1: {10: {100: '111'}, 20: '222'}, 2: '22'}
        cd = CastDict(d, cast_map={str: int})
        self.assertDictEqual(cd, {1: {10: {100: 111}, 20: 222}, 2: 22})
        self.assertIsInstance(cd[1], CastDict)
        self.assertIsInstance(cd[1][10], CastDict)

    def test_cast8(self):
        d = {1: [{2: '20'}, {3: '30'}]}
        cd = CastDict(d, cast_map={str: int})
        self.assertDictEqual(cd, {1: [{2: 20}, {3: 30}]})
        self.assertIsInstance(cd[1], CastList)
        self.assertIsInstance(cd[1][0], CastDict)

    def test_cast9(self):
        d = {1: (1, 2, 3)}
        cd = CastDict(d, cast_map={tuple: sum})
        self.assertDictEqual(cd, {1: 6})

    def test_cast10(self):
        d = {1: ('a', 'b', 'c'), 2: ('1', '2', '3')}
        cd = CastDict(d, cast_map={tuple: lambda x: ':'.join(x)})
        self.assertDictEqual(cd, {1: 'a:b:c', 2: '1:2:3'})




"""
Run via python -m unittest in parent folder

"""

from unittest import TestCase

from empty import *


class TestEmptyDict(TestCase):

    def test_null_init(self):
        self.assertEqual(EmptyDict(), {})

    def test_init(self):
        data = {1: 10, 2: 20, 3:30}
        e = EmptyDict(data)
        self.assertEqual(e, {})

    def test_retreive(self):
        data = {'a': 'A', 'B': 'B', 'c': 'C'}
        e = EmptyDict(data)
        self.assertRaises(KeyError, lambda x: e[x], 'a')
        self.assertIsNone(e.get('a'))

    def test_len(self):
        data = {'a': 'A', 'B': 'B', 'c': 'C'}
        e = EmptyDict(data)
        self.assertEqual(len(e), 0)

    def test_iter(self):
        data = {'a': 'A', 'B': 'B', 'c': 'C'}
        e = EmptyDict(data)
        self.assertListEqual(list(iter(e)), [])
        self.assertListEqual(list(e.keys()), [])
        self.assertListEqual(list(e.values()), [])
        self.assertListEqual(list(e.items()), [])

    def test_add_item_(self):
        e = EmptyDict()
        e['x'] = 'X'
        self.assertFalse('x' in e)
        self.assertIsNone(e.get('x'))
        self.assertRaises(KeyError, lambda e, x: e[x], e, 'x')

    def test_del_item(self):
        data = {'a': 'A', 'B': 'B', 'c': 'C'}
        e = EmptyDict(data)
        def f(e, x):
            del e[x]
        self.assertRaises(KeyError, f, e, 'a')
